from __future__ import absolute_import, division, print_function

from koala.CellBase import CellBase
from koala.ExcelError import ErrorCodes, ExcelError
from koala.utils import *

from openpyxl.compat import unicode


# WARNING: Range should never be imported directly. Import Range from excelutils instead.

### Range Utils ###

parse_cell_addr_cache = {}


def parse_cell_address(ref):
    # A1 => (1, 'A')
    try:
        if ref not in parse_cell_addr_cache:
            found = re.search(CELL_REF_RE, ref)
            col = found.group(1)
            row = found.group(2)
            result = (int(row), col)
            parse_cell_addr_cache[ref] = result
            return result
        else:
            return parse_cell_addr_cache[ref]
    except:
        raise Exception('Couldn\'t find match in cell ref')

get_cell_addr_cache = {}


def get_cell_address(sheet, tuple):
    # Sheet1, (1, 'A') => Sheet1!A1

    if (sheet, tuple) not in get_cell_addr_cache:
        row = tuple[0]
        col = tuple[1]

        if sheet is not None:
            addr = sheet + '!' + str(col) + str(row)
            get_cell_addr_cache[(sheet, tuple)] = addr
            return addr
        else:
            addr = col + str(row)
            get_cell_addr_cache[(sheet, tuple)] = addr
            return addr
    else:
        return get_cell_addr_cache[(sheet, tuple)]


def check_value(a):
    if isinstance(a, ExcelError):
        return a
    elif isinstance(a, str) and a in ErrorCodes:
        return ExcelError(a)

    try:  # This is to avoid None or Exception returned by Range operations
        if float(a) or isinstance(a, (unicode, str)):
            return a
        else:
            return 0
    except:
        if a == 'True':
            return True
        elif a == 'False':
            return False
        else:
            return 0


class RangeCore(dict):

    def __init__(
            self, reference,
            values=None, cellmap=None, nrows=None, ncols=None,
            name=None):

        is_pointer = False

        if type(reference) == dict:
            is_pointer = True

            name = '%s:%s' % (reference['start'], reference['end'])

        self.__pointer = is_pointer
        self.__reference = reference
        self.__cellmap = cellmap

        # These get redefined in `__build()`
        self.__name = name
        self.__origin = None
        self.__addresses = []
        self.__order = []
        self.__length = None
        self.__nrows = None
        self.__ncols = None
        self.__type = None
        self.__sheet = None
        self.__empty = None

        if not is_pointer:
            self.__build(
                reference=reference, values=values,
                cellmap=cellmap, nrows=nrows, ncols=ncols, name=name)

    # A separate function from __init__ is necessary so that it can be called from outside
    def __build(
            self, reference,
            values=None, cellmap=None, nrows=None, ncols=None,
            name=None, debug=False):

        if type(reference) == list:  # some Range calculations such as excellib.countifs() use filtered keys
            cells = reference
        else:
            reference = reference.replace('$', '')
            try:
                cells, nrows, ncols = resolve_range(
                    reference, should_flatten=True)
            except Exception as e:
                print('Pb with ref', reference, e)
                return ValueError('Range ERROR')  # Will still be considered as a Range object, since we are inside __init__...

        origin = parse_cell_address(cells[0]) if len(cells) > 0 else None  # origin of Range

        if cellmap:
            cells = [cell for cell in cells if cell in cellmap]

        if values:
            if len(cells) != len(values):
                raise Exception(
                    "Cells and values in a Range must have the same size",
                    reference)

        result = []
        order = []

        empty = True

        for index, cell in enumerate(cells):
            row, col = parse_cell_address(cell)
            order.append((row, col))
            try:
                if cellmap:
                    result.append(((row, col), cellmap[cell]))
                    empty = empty and cellmap[cell].value is None
                else:
                    if isinstance(values[index], RangeCore):
                        raise Exception(
                            'Range can\'t be values of Range',
                            reference)
                    result.append(((row, col), values[index]))
                    empty = empty and values[index] is None

            except:  # when you don't provide any values
                result.append(((row, col), None))

        try:
            sheet = cells[0].split('!')[0]
        except:
            sheet = None

        # dont allow messing with these params
        if type(reference) == list:
            self.__name = name
        elif not self.is_pointer:  # when building pointers, name shouldn't be updated, but in that case reference is not a dict
            self.__name = reference
        else:
            print('Pb with Name', reference, name)
        self.__origin = origin
        self.__addresses = cells
        self.__order = order
        self.__length = len(cells)
        self.__nrows = nrows
        self.__ncols = ncols
        self.__empty = empty

        if ncols == 1 and nrows == 1:
            self.__type = 'scalar'
        elif ncols == 1:
            self.__type = 'vertical'
        elif nrows == 1:
            self.__type = 'horizontal'
        else:
            self.__type = 'bidimensional'
        self.__sheet = sheet

        dict.__init__(self, result)

    def build(
            self, reference=None, values=None,
            nrows=None, ncols=None, name=None, debug=False):
        self.__build(
            reference=reference, values=values,
            cellmap=self.__cellmap,
            nrows=nrows, ncols=ncols, name=name, debug=debug)

    @property
    def reference(self):
        return self.__reference

    @property
    def name(self):
        return self.__name

    @property
    def origin(self):
        return self.__origin

    @property
    def addresses(self):
        return self.__addresses

    @property
    def order(self):
        return self.__order

    @property
    def length(self):
        return self.__length

    @property
    def is_pointer(self):
        return self.__pointer

    @property
    def nrows(self):
        return self.__nrows

    @property
    def ncols(self):
        return self.__ncols

    @property
    def is_empty(self):
        return self.__empty

    @property
    def type(self):
        return self.__type

    @property
    def sheet(self):
        return self.__sheet

    @property
    def values(self):
        if self.__cellmap:
            values = []
            for cell in self.cells:
                values.append(cell.value)
            return values
        else:
            return self.cells

    @values.setter
    def values(self, new_values):
        if self.__cellmap:
            for index, cell in enumerate(self.cells):
                if index < len(new_values):
                    cell.value = new_values[index]
        else:
            for index, value in enumerate(self.order):
                if index < len(new_values):
                    self[value] = new_values[index]

    @property
    def cells(self):
        return [self[c] for c in self.order]

    def get(self, row, col=None):
        nr = self.nrows
        nc = self.ncols

        values = self.values
        cells = self.addresses

        if nr == 1 or nc == 1:  # 1-dim range
            if col is not None:
                raise Exception(
                    'Trying to access 1-dim range value with 2 coordinates')
            else:
                return values[row - 1]

        else:  # could be optimised
            origin_col = col2num(self.origin[1])
            origin_row = self.origin[0]

            if row == 0:  # get column

                out_col = num2col(int(col2num(self.origin[1]) + col - 1))

                tuples = [
                    (r, out_col)
                    for r in range(origin_row, origin_row + self.nrows)
                ]

                cells = []
                values = []

                for t in tuples:
                    if t in self:
                        values.append(self[t].value)
                    else:
                        values.append(None)
                    cells.append(get_cell_address(self.sheet, t))

                return RangeCore(
                    cells, values=values, nrows=len(cells), ncols=1)

            elif col == 0:  # get row

                out_row = self.origin[0] + row - 1

                tuples = [
                    (out_row, c)
                    for c in range(origin_col, origin_col + self.ncols)
                ]

                cells = []
                values = []

                for t in tuples:
                    if t in self:
                        values.append(self[t].value)
                    else:
                        values.append(None)
                    cells.append(get_cell_address(self.sheet, t))

                return RangeCore(
                    cells, values=values, nrows=1, ncols=len(cells))

            else:
                base_col_number = col2num(cells[0][0])
                new_ref = num2col(col + base_col_number - 1) + str(row)
                new_value = values[(row - 1) * nc + (col - 1)]

                return new_value

    @staticmethod
    def filter(range, bool_range):

        if range.type == 'bidimensional':
            raise Exception('Cant use filter on bidimensional Ranges')

        filtered_addresses = []
        filtered_values = []

        for index, value in enumerate(range.values):
            test_value = bool_range.values[index]

            if type(test_value) != bool:
                raise Exception(
                    'RangeCore.filter must be used with bool Range as a second argument')

            if test_value:
                filtered_addresses.append(range.addresses[index])
                filtered_values.append(value)

        ncols = 1
        nrows = 1

        if range.type == 'vertical':
            nrows = len(filtered_values)
        elif range.type == 'horizontal':
            ncols = len(filtered_values)

        return RangeCore(
            filtered_addresses, filtered_values, nrows=nrows, ncols=ncols)

    @staticmethod
    def find_associated_cell(ref, range):
        # This function retrieves the cell associated to ref in a Range
        # For instance, in the range [A1, B1, C1], the cell associated to B2 is B1
        # This is useful to mimic the way Excel works

        if ref is not None:
            row, col = ref

            if (range.length) == 0:  # if a Range is empty, it means normally that all its cells are empty
                return None
            elif range.type == "vertical":
                if (row, range.origin[1]) in range.order:
                    return range.addresses[
                        range.order.index((row, range.origin[1]))
                    ]
                else:
                    return None
            elif range.type == "horizontal":
                if (range.origin[0], col) in range.order:
                    return range.addresses[
                        range.order.index((range.origin[0], col))
                    ]
                else:
                    return None
            elif range.type == "scalar":
                if (row, range.origin[1]) in range.order:
                    return range.addresses[
                        range.order.index((row, range.origin[1]))
                    ]
                elif (range.origin[0], col) in range.order:
                    return range.addresses[
                        range.order.index((range.origin[0], col))
                    ]
                elif (row, col) in range.order:
                    return range.addresses[
                        range.order.index((row, col))
                    ]
                else:
                    return None

            else:
                return None
        else:
            return None

    @staticmethod
    def find_associated_value(ref, item):
        # This function is ALMOST equivalent to RangeCore.find_associated_cell, but retrieves the value and not the Cell.
        row, col = ref

        if isinstance(item, RangeCore):
            try:
                if (item.length) == 0:  # if a Range is empty, it means normally that all its cells are empty
                    item_value = 0
                elif item.type == "vertical":
                    if item.__cellmap is not None:
                        item_value = item[(row, item.origin[1])].value if (row, item.origin[1]) in item else None
                    else:
                        item_value = item[(row, item.origin[1])] if (row, item.origin[1]) in item else None
                elif item.type == "horizontal":

                    if item.__cellmap is not None:
                        try:
                            item_value = item[(item.origin[0], col)].value if (item.origin[0], col) in item else None
                        except:
                            raise Exception
                    else:
                        item_value = item[(item.origin[0], col)] if (item.origin[0], col) in item else None
                else:
                    raise ExcelError('#VALUE!', 'cannot use find_associated_value on %s' % item.type)
            except ExcelError as e:
                raise Exception('First argument of Range operation is not valid: ' + e.value)
        elif item is None:
            item_value = 0
        else:
            item_value = item

        return item_value

    @staticmethod
    def apply(func, first, second, ref=None):
        # This function decides whether RangeCore.apply_one or RangeCore.apply_all should be used
        # This is a necessary complement to what is decided in graph.py:OperandNode.emit()

        is_associated = False

        if ref:
            if isinstance(first, RangeCore):
                if first.length == 0:
                    first = 0
                else:
                    is_associated = RangeCore.find_associated_cell(ref, first) is not None
            elif second is not None and isinstance(second, RangeCore):
                if second.length == 0:
                    second = 0
                else:
                    is_associated = RangeCore.find_associated_cell(ref, second) is not None
            else:
                is_associated = False

            if is_associated:
                return RangeCore.apply_one(func, first, second, ref)
            else:
                return RangeCore.apply_all(func, first, second, ref)
        else:
            return RangeCore.apply_all(func, first, second, ref)

    @staticmethod
    def apply_one(func, first, second, ref=None):
        # This function applies a function to range operands, only for the cells associated to ref
        # Note that non-range operands are supported by RangeCore.find_associated_value()

        function = func_dict[func]

        if ref is None:
            first_value = first
            second_value = second
        else:
            first_value = RangeCore.find_associated_value(ref, first)
            second_value = RangeCore.find_associated_value(ref, second)

        return function(first_value, second_value)

    @staticmethod
    def apply_all(func, first, second, ref=None):
        # This function applies a function to range operands, for all the cells in the Ranges

        function = func_dict[func]

        # Here, the first arg of RangeCore() has little importance: TBC
        if isinstance(first, RangeCore) and isinstance(second, RangeCore):
            if first.length != second.length:
                raise ExcelError('#VALUE!', 'apply_all must have 2 Ranges of identical length')

            vals = [function(
                x.value if isinstance(x, CellBase) else x,
                y.value if isinstance(x, CellBase) else y
            ) for x, y in zip(first.cells, second.cells)]

            return RangeCore(
                first.addresses, vals, nrows=first.nrows, ncols=first.ncols)

        elif isinstance(first, RangeCore):
            vals = [function(
                x.value if isinstance(x, CellBase) else x,
                second
            ) for x in first.cells]

            return RangeCore(
                first.addresses, vals, nrows=first.nrows, ncols=first.ncols)

        elif isinstance(second, RangeCore):
            vals = [function(
                first,
                x.value if isinstance(x, CellBase) else x
            ) for x in second.cells]

            return RangeCore(
                second.addresses, vals, nrows=second.nrows, ncols=second.ncols)

        else:
            return function(first, second)

    @staticmethod
    def add(a, b):
        try:
            return check_value(a) + check_value(b)
        except Exception as e:
            return ExcelError('#N/A', e)

    @staticmethod
    def substract(a, b):
        try:
            return check_value(a) - check_value(b)
        except Exception as e:
            return ExcelError('#N/A', e)

    @staticmethod
    def minus(a, b=None):
        # b is not used, but needed in the signature. Maybe could be better
        try:
            return -check_value(a)
        except Exception as e:
            return ExcelError('#N/A', e)

    @staticmethod
    def multiply(a, b):
        try:
            return check_value(a) * check_value(b)
        except Exception as e:
            return ExcelError('#N/A', e)

    @staticmethod
    def divide(a, b):
        try:
            return old_div(float(check_value(a)), float(check_value(b)))
        except Exception as e:
            return ExcelError('#DIV/0!', e)

    @staticmethod
    def is_equal(a, b):
        try:
            if not isinstance(a, (str, unicode)):
                a = check_value(a)
            if not isinstance(b, (str, unicode)):
                b = check_value(b)

            return is_almost_equal(a, b, precision=0.00001)
        except Exception as e:
            return ExcelError('#N/A', e)

    @staticmethod
    def is_not_equal(a, b):
        try:
            if not isinstance(a, (str, unicode)):
                a = check_value(a)
            if not isinstance(a, (str, unicode)):
                b = check_value(b)

            return a != b
        except Exception as e:
            return ExcelError('#N/A', e)

    @staticmethod
    def is_strictly_superior(a, b):
        try:
            return check_value(a) > check_value(b)
        except Exception as e:
            return ExcelError('#N/A', e)

    @staticmethod
    def is_strictly_inferior(a, b):
        try:
            return check_value(a) < check_value(b)
        except Exception as e:
            return ExcelError('#N/A', e)

    @staticmethod
    def is_superior_or_equal(a, b):
        try:
            a = check_value(a)
            b = check_value(b)

            return a > b or is_almost_equal(a, b)
        except Exception as e:
            return ExcelError('#N/A', e)

    @staticmethod
    def is_inferior_or_equal(a, b):
        try:
            a = check_value(a)
            b = check_value(b)

            return a < b or is_almost_equal(a, b)
        except Exception as e:
            return ExcelError('#N/A', e)

func_dict = {
    "multiply": RangeCore.multiply,
    "divide": RangeCore.divide,
    "add": RangeCore.add,
    "substract": RangeCore.substract,
    "minus": RangeCore.minus,
    "is_equal": RangeCore.is_equal,
    "is_not_equal": RangeCore.is_not_equal,
    "is_strictly_superior": RangeCore.is_strictly_superior,
    "is_strictly_inferior": RangeCore.is_strictly_inferior,
    "is_superior_or_equal": RangeCore.is_superior_or_equal,
    "is_inferior_or_equal": RangeCore.is_inferior_or_equal,
}


def RangeFactory(cellmap=None):

    class Range(RangeCore):

        def __init__(self, reference, values=None):
            super(Range, self).__init__(reference, values, cellmap=cellmap)

    return Range



"""Setup script."""

from setuptools import setup, find_packages


if __name__ == '__main__':
    setup(
        name="koala2",

        version="0.0.30",

        author="Ants, open innovation lab",
        author_email="contact@weareants.fr",

        packages=find_packages(),

        include_package_data=True,

        url="https://github.com/anthill/koala",

        license="GNU GPL3",
        description=(
            "A python module to extract all the content of an Excel document "
            "and enable calculation without Excel"
        ),

        long_description=open("README.md").read(),

        install_requires=[
            'networkx >= 2.1',
            'openpyxl >= 2.5.3',
            'numpy >= 1.14.2',
            'Cython >= 0.28.2',
            'lxml >= 4.1.1',
            'six >= 1.11.0',
        ]
    )



import os.path
import sys
import unittest

from koala.reader import read_archive, read_cells
from koala import ExcelCompiler, Spreadsheet


# # This fails, needs to be adapted
class Test_SharedFormula(unittest.TestCase):

    def setUp(self):
        file_name = os.path.abspath("./tests/files/SharedFormula.xlsx")
        archive = read_archive(file_name)
        self.cells = read_cells(archive)

    @unittest.skip('This test fails.')
    def test_nb_formulas(self):
        self.assertEqual(len([ref_cell for ref_cell in list(self.cells.items()) if ref_cell[1].formula is not None]), 13)

    def test_shared_formulas_content(self):
        self.assertEqual(self.cells[('Shared_formula!G2')].formula, 'G1 + 10 * L1 + $A$1')

    def test_text_content(self):
        self.assertEqual(self.cells[('Shared_formula!C12')].value, 'Romain')

    @unittest.skip('This test fails.')
    def test_types(self):
        nb_int = len([ref_cell1 for ref_cell1 in list(self.cells.items()) if type(ref_cell1[1].value) == int])
        nb_float = len([ref_cell2 for ref_cell2 in list(self.cells.items()) if type(ref_cell2[1].value) == float])
        nb_bool = len([ref_cell3 for ref_cell3 in list(self.cells.items()) if type(ref_cell3[1].value) == bool])
        nb_str = len([ref_cell4 for ref_cell4 in list(self.cells.items()) if type(ref_cell4[1].value) == str])

        self.assertTrue(nb_int == 21 and nb_float == 3 and nb_bool == 2 and nb_str == 10)


class Test_NamedRanges(unittest.TestCase):


    def setUp(self):
        c = ExcelCompiler("./tests/files/NamedRanges.xlsx", ignore_sheets = ['IHS'])
        self.graph = c.gen_graph()
        sys.setrecursionlimit(10000)

    def test_before_set_value(self):
        self.assertTrue(self.graph.evaluate('INPUT') == 1 and self.graph.evaluate('Sheet1!A1') == 1 and self.graph.evaluate('RESULT') == 187)

    @unittest.skip('This test fails.')
    def test_after_set_value(self):
        self.graph.set_value('INPUT', 2025)
        self.assertTrue(self.graph.evaluate('INPUT') == 2025 and self.graph.evaluate('Sheet1!A1') == 2025 and self.graph.evaluate('RESULT') == 2211)


class Test_DumpDict(unittest.TestCase):

    def setUp(self):
        c = ExcelCompiler("./tests/files/NamedRanges.xlsx", ignore_sheets=['IHS'])
        self.graph = c.gen_graph()
        sys.setrecursionlimit(10000)

    def test_no_set_value(self):
        graph = self.graph
        self.assertTrue(graph.evaluate('INPUT') == 1)
        self.assertTrue(graph.evaluate('Sheet1!A1') == 1)
        self.assertTrue(graph.evaluate('RESULT') == 187)

    def test_set_value(self):
        # Clone object
        data = self.graph.asdict()
        graph = Spreadsheet.from_dict(data)

        # Set value and check result in clone.
        graph.set_value('INPUT', 2025)
        self.assertTrue(graph.evaluate('INPUT') == 2025)
        self.assertTrue(graph.evaluate('Sheet1!A1') == 2025)
        # self.assertTrue(graph.evaluate('RESULT') == 2211)

        # Check original not changed.
        graph = self.graph
        self.assertTrue(graph.evaluate('INPUT') == 1)
        self.assertTrue(graph.evaluate('Sheet1!A1') == 1)
        self.assertTrue(graph.evaluate('RESULT') == 187)


class Test_Dump(unittest.TestCase):


    def setUp(self):
        c = ExcelCompiler("./tests/files/NamedRanges.xlsx", ignore_sheets = ['IHS'])
        self.graph = c.gen_graph()
        sys.setrecursionlimit(10000)

    def test_no_set_value(self):
        graph = self.graph
        self.assertTrue(graph.evaluate('INPUT') == 1)
        self.assertTrue(graph.evaluate('Sheet1!A1') == 1)
        self.assertTrue(graph.evaluate('RESULT') == 187)

    def test_set_value(self):
        # Clone object
        self.graph.dump("dump.txt.gz")
        graph = Spreadsheet.load("dump.txt.gz")

        # Set value and check result in clone.
        graph.set_value('INPUT', 2025)
        self.assertTrue(graph.evaluate('INPUT') == 2025)
        self.assertTrue(graph.evaluate('Sheet1!A1') == 2025)
        # self.assertTrue(graph.evaluate('RESULT') == 2211)

        # Check original not changed.
        graph = self.graph
        self.assertTrue(graph.evaluate('INPUT') == 1)
        self.assertTrue(graph.evaluate('Sheet1!A1') == 1)
        self.assertTrue(graph.evaluate('RESULT') == 187)


class Test_DumpJson(unittest.TestCase):


    def setUp(self):
        c = ExcelCompiler("./tests/files/NamedRanges.xlsx", ignore_sheets = ['IHS'])
        self.graph = c.gen_graph()
        sys.setrecursionlimit(10000)

    def test_no_set_value(self):
        graph = self.graph
        self.assertTrue(graph.evaluate('INPUT') == 1)
        self.assertTrue(graph.evaluate('Sheet1!A1') == 1)
        self.assertTrue(graph.evaluate('RESULT') == 187)

    def test_set_value(self):
        # Clone object
        self.graph.dump_json("dump.txt.gz")
        graph = Spreadsheet.load_json("dump.txt.gz")

        # Set value and check result in clone.
        graph.set_value('INPUT', 2025)
        self.assertTrue(graph.evaluate('INPUT') == 2025)
        self.assertTrue(graph.evaluate('Sheet1!A1') == 2025)
        # self.assertTrue(graph.evaluate('RESULT') == 2211)

        # Check original not changed.
        graph = self.graph
        self.assertTrue(graph.evaluate('INPUT') == 1)
        self.assertTrue(graph.evaluate('Sheet1!A1') == 1)
        self.assertTrue(graph.evaluate('RESULT') == 187)



from __future__ import absolute_import

import pyximport; pyximport.install()

import unittest

from koala.excellib import *


class Test_VDB(unittest.TestCase):
    def test_vdb_basic(self):
        cost = 575000
        salvage = 5000
        life = 10
        rate = 1.5
        start = 3
        end = 5

        obj = 102160.546875

        self.assertEqual(vdb(cost, salvage, life, start, end, rate), obj)

    def test_vdb_partial(self):
        cost = 1
        salvage = 0
        life = 14
        rate = 1.25
        start = 11.5
        end = 12.5

        obj = 0.068726290454684

        self.assertEqual(round(vdb(cost, salvage, life, start, end, rate), 15), obj)

    def test_vdb_partial_no_switch(self):
        cost = 1
        salvage = 0
        life = 5.0
        rate = 2.5
        start = 0.5
        end = 1.5

        obj = 0.375

        self.assertEqual(vdb(cost, salvage, life, start, end, rate, True), obj)


class Test_SLN(unittest.TestCase):
    def test_sln_basic(self):
        self.assertEqual(sln(30000, 5000, 10), 2500)


class Test_Choose(unittest.TestCase):
    def test_choose_basic(self):
        self.assertEqual(choose(3, 'John', 'Paul', 'George', 'Ringo'), 'George')

    def test_choose_fraction(self):
        self.assertEqual(choose(3.4, 'John', 'Paul', 'George', 'Ringo'), 'George')

    def test_choose_incorrect_index(self):
        self.assertEqual(type(choose(3, 2)), ExcelError)


class Test_Irr(unittest.TestCase):
    def test_irr_basic(self):
        self.assertEqual(round(irr([-100, 39, 59, 55, 20], 0), 7), 0.2809484)

    def test_irr_with_guess_non_null(self):
        with self.assertRaises(ValueError):
            irr([-100, 39, 59, 55, 20], 2)


class Test_Offset(unittest.TestCase):
    @unittest.skip('This test fails.')
    def test_offset_height_not_integer(self):
        with self.assertRaises(ExcelError):
            offset('Sheet1!A1', 'e', 2)

    @unittest.skip('This test fails.')
    def test_offset_height_is_zero(self):
        with self.assertRaises(ExcelError):
            offset('Sheet1!A1', 1, 2, 0, 1)

    @unittest.skip('This test fails.')
    def test_offset_only_height(self):
        with self.assertRaises(ExcelError):
            offset('Sheet1!A1', 1)

    @unittest.skip('This test fails.')
    def test_offset_out_of_bounds(self):
        with self.assertRaises(ExcelError):
            offset('Sheet1!A1', 1, -2)

    def test_offset_regular(self):
        self.assertEqual(offset('A1:B2', 1, 2), 'C2')

    def test_offset_with_sheet(self):
        self.assertEqual(offset('Sheet1!A1:Sheet1!B2', 1, 2), 'Sheet1!C2')

    def test_offset_rectangular(self):
        self.assertEqual(offset('Sheet1!A1:B2', 1, 2, 2, 3), 'Sheet1!C2:E3')


class Test_Lookup(unittest.TestCase):
    def test_lookup_with_result_range(self):
        range1 = Range('A1:A3', [1, 2, 3])
        range2 = Range('B1:B3', ['blue', 'orange', 'green'])

        self.assertEqual(lookup(2, range1, range2), 'orange')

    def test_lookup_find_closest_inferior(self):
        range = Range('A1:A3', [1, 2, 3])
        self.assertEqual(lookup(2.5, range), 2)

    def test_lookup_basic(self):
        range = Range('A1:A3', [1, 2, 3])
        self.assertEqual(lookup(2, range), 2)


class Test_Average(unittest.TestCase):
    def test_average(self):
        range = Range('A1:A3', [2, 4, 6])
        value = 8
        self.assertEqual(average(range, value), 5)


class Test_Min(unittest.TestCase):
    def test_min_range_and_value(self):
        range = Range('A1:A3', [1, 23, 3])
        value = 20
        self.assertEqual(xmin(range, value), 1)


class Test_Max(unittest.TestCase):
    def test_max_range_and_value(self):
        range = Range('A1:A3', [1, 23, 3])
        value = 20
        self.assertEqual(xmax(range, value), 23)


class Test_Sum(unittest.TestCase):
    def test_sum_ignores_non_numeric(self):
        range = Range('A1:A3', [1, 'e', 3])
        self.assertEqual(xsum(range), 4)

    def test_sum_returns_zero_when_no_numeric(self):
        range = Range('A1:A3', ['ER', 'Er', 're'])
        value = 'ererr'
        self.assertEqual(xsum(range, value), 0)

    def test_sum_excludes_booleans_from_nested_ranges(self):
        range = Range('A1:A3', [True, 2, 1])
        value = True
        self.assertEqual(xsum(range, value), 4)

    def test_sum_range_and_value(self):
        range = Range('A1:A3', [1, 2, 3])
        value = 20
        self.assertEqual(xsum(range, value), 26)


class Test_Iferror(unittest.TestCase): # Need rewriting
    @unittest.skip('This test fails.')
    def test_when_error(self):
        self.assertEqual(iferror('3 / 0', 4), 4)

    @unittest.skip('This test fails.')
    def test_when_no_error(self):
        self.assertEqual(iferror('3 * 10', 4), 30)


class Test_Index(unittest.TestCase):
    @unittest.skip('This test fails.')
    def test_index_not_range(self):
        with self.assertRaises(ExcelError):
            index([0, 0.2, 0.15], 2, 4)

    @unittest.skip('This test fails.')
    def test_index_row_not_number(self):
        with self.assertRaises(ExcelError):
            index(resolve_range('A1:A3'), 'e', 1)

    @unittest.skip('This test fails.')
    def test_index_col_not_number(self):
        with self.assertRaises(ExcelError):
            index(resolve_range('A1:D3'), 1, 'e')

    @unittest.skip('This test fails.')
    def test_index_dim_zero(self):
        with self.assertRaises(ExcelError):
            index(resolve_range('A1:D3'), 0, 0)

    def test_index_1_dim_2_coords(self):
        self.assertEqual(index(resolve_range('A1:A3'), 3, 1), 'A3')

    @unittest.skip('This test fails.')
    def test_index_1_dim_out_of_range(self):
        with self.assertRaises(ExcelError):
            index(resolve_range('A1:A3'), 4)

    def test_index_1_dim(self):
        self.assertEqual(index(resolve_range('A1:A3'), 3), 'A3')

    @unittest.skip('This test fails.')
    def test_index_2_dim_1_coord(self):
        with self.assertRaises(ExcelError):
            index(resolve_range('D1:F2'), 2)

    @unittest.skip('This test fails.')
    def test_index_2_dim_out_of_range(self):
        with self.assertRaises(ExcelError):
            index(resolve_range('D1:F2'), 2, 6)

    def test_index_2_dim_row_0(self):
        self.assertEqual(index(resolve_range('D1:F2'), 0, 3), ['F1', 'F2'])

    def test_index_2_dim_col_0(self):
        self.assertEqual(index(resolve_range('D1:F2'), 2, 0), ['D2', 'E2', 'F2'])

    @unittest.skip('This test fails.')
    def test_index_2_dim_col_0_ref_not_found(self):
        # range = Range(['D1', 'E1', 'F1', 'D2', 'E2', 'F2'], [1, 2, 3, 4, 5, 6])

        with self.assertRaises(ExcelError):
            index(resolve_range('D1:F2'), 2, 0)

    def test_index_2_dim(self):
        self.assertEqual(index(resolve_range('D1:F2'), 2, 2), 'E2')


class Test_SumProduct(unittest.TestCase):
    @unittest.skip('This test fails.')
    def test_ranges_with_different_sizes(self):
        range1 = Range('A1:A3', [1, 10, 3])
        range2 = Range('B1:B4', [3, 3, 1, 2])

        with self.assertRaises(ExcelError):
            sumproduct(range1, range2)

    def test_regular(self):
        range1 = Range('A1:A3', [1, 10, 3])
        range2 = Range('B1:B3', [3, 3, 1])

        self.assertEqual(sumproduct(range1, range2), 36)


class Test_SumIf(unittest.TestCase):
    @unittest.skip('This test fails.')
    def test_range_is_a_list(self):
        with self.assertRaises(ExcelError):
            sumif(12, 12)

    @unittest.skip('This test fails.')
    def test_sum_range_is_a_list(self):
        with self.assertRaises(ExcelError):
            sumif(12, 12, 12)

    def test_criteria_is_not_number_string_boolean(self):
        range1 = Range('A1:A3', [1, 2, 3])
        range2 = Range('A1:A2', [1, 2])

        self.assertEqual(sumif(range1, range2), 0)

    def test_regular_with_number_criteria(self):
        range = Range('A1:A5', [1, 1, 2, 2, 2])

        self.assertEqual(sumif(range, 2), 6)

    def test_regular_with_string_criteria(self):
        range = Range('A1:A6', [1, 2, 3, 4, 5, 6])

        self.assertEqual(sumif(range, ">=3"), 18)

    def test_sum_range(self):
        range1 = Range('A1:A5', [1, 2, 3, 4, 5])
        range2 = Range('A1:A5', [100, 123, 12, 23, 633])

        self.assertEqual(sumif(range1, ">=3", range2), 668)

    def test_sum_range_with_more_indexes(self):
        range1 = Range('A1:A5', [1, 2, 3, 4, 5])
        range2 = Range('A1:A6', [100, 123, 12, 23, 633, 1])

        self.assertEqual(sumif(range1, ">=3", range2), 668)

    def test_sum_range_with_less_indexes(self):
        range1 = Range('A1:A5', [1, 2, 3, 4, 5])
        range2 = Range('A1:A4', [100, 123, 12, 23])

        self.assertEqual(sumif(range1, ">=3", range2), 35)


class Test_IsNa(unittest.TestCase):
    # This function might need more solid testing

    @unittest.skip('This test fails.')
    def test_isNa_false(self):
        self.assertFalse(isNa('2 + 1'))

    @unittest.skip('This test fails.')
    def test_isNa_true(self):
        self.assertTrue(isNa('x + 1'))


class Test_Yearfrac(unittest.TestCase):
    def test_start_date_must_be_number(self):
        self.assertEqual(type(yearfrac('not a number', 1)), ExcelError)

    def test_end_date_must_be_number(self):
        self.assertEqual(type(yearfrac(1, 'not a number')), ExcelError)

    def test_start_date_must_be_positive(self):
        self.assertEqual(type(yearfrac(-1, 0)), ExcelError)

    def test_end_date_must_be_positive(self):
        self.assertEqual(type(yearfrac(0, -1)), ExcelError)

    @unittest.skip('This test fails.')
    def test_basis_must_be_between_0_and_4(self):
        with self.assertRaises(ExcelError):
            yearfrac(1, 2, 5)

    def test_yearfrac_basis_0(self):
        self.assertAlmostEqual(yearfrac(date(2008, 1, 1), date(2015, 4, 20)), 7.30277777777778)

    def test_yearfrac_basis_1(self):
        self.assertAlmostEqual(yearfrac(date(2008, 1, 1), date(2015, 4, 20), 1), 7.299110198)

    def test_yearfrac_basis_2(self):
        self.assertAlmostEqual(yearfrac(date(2008, 1, 1), date(2015, 4, 20), 2), 7.405555556)

    def test_yearfrac_basis_3(self):
        self.assertAlmostEqual(yearfrac(date(2008, 1, 1), date(2015, 4, 20), 3), 7.304109589)

    def test_yearfrac_basis_4(self):
        self.assertAlmostEqual(yearfrac(date(2008, 1, 1), date(2015, 4, 20), 4), 7.302777778)

    def test_yearfrac_inverted(self):
        self.assertAlmostEqual(yearfrac(date(2015, 4, 20), date(2008, 1, 1)), yearfrac(date(2008, 1, 1), date(2015, 4, 20)))


class Test_Date(unittest.TestCase):
    @unittest.skip('This test fails.')
    def test_year_must_be_integer(self):
        with self.assertRaises(ExcelError):
            date('2016', 1, 1)

    @unittest.skip('This test fails.')
    def test_month_must_be_integer(self):
        with self.assertRaises(ExcelError):
            date(2016, '1', 1)

    @unittest.skip('This test fails.')
    def test_day_must_be_integer(self):
        with self.assertRaises(ExcelError):
            date(2016, 1, '1')

    @unittest.skip('This test fails.')
    def test_year_must_be_positive(self):
        with self.assertRaises(ExcelError):
            date(-1, 1, 1)

    @unittest.skip('This test fails.')
    def test_year_must_have_less_than_10000(self):
        with self.assertRaises(ExcelError):
            date(10000, 1, 1)

    @unittest.skip('This test fails.')
    def test_result_must_be_positive(self):
        with self.assertRaises(ArithmeticError):
            date(1900, 1, -1)

    def test_not_stricly_positive_month_substracts(self):
        self.assertEqual(date(2009, -1, 1), date(2008, 11, 1))

    def test_not_stricly_positive_day_substracts(self):
        self.assertEqual(date(2009, 1, -1), date(2008, 12, 30))

    def test_month_superior_to_12_change_year(self):
        self.assertEqual(date(2009, 14, 1), date(2010, 2, 1))

    def test_day_superior_to_365_change_year(self):
        self.assertEqual(date(2009, 1, 400), date(2010, 2, 4))

    def test_year_for_29_feb(self):
        self.assertEqual(date(2008, 2, 29), 39507)

    def test_year_regular(self):
        self.assertEqual(date(2008, 11, 3), 39755)

class Test_Mid(unittest.TestCase):
    @unittest.skip('This test fails.')
    def test_start_num_must_be_integer(self):
        with self.assertRaises(ExcelError):
            mid('Romain', 1.1, 2)

    @unittest.skip('This test fails.')
    def test_num_chars_must_be_integer(self):
        with self.assertRaises(ExcelError):
            mid('Romain', 1, 2.1)

    @unittest.skip('This test fails.')
    def test_start_num_must_be_superior_or_equal_to_1(self):
        with self.assertRaises(ExcelError):
            mid('Romain', 0, 3)

    @unittest.skip('This test fails.')
    def test_num_chars_must_be_positive(self):
        with self.assertRaises(ExcelError):
            mid('Romain', 1, -1)

    def test_mid(self):
        self.assertEqual(mid('Romain', 2, 9), 'main')


class Test_Round(unittest.TestCase):
    @unittest.skip('This test fails.')
    def test_nb_must_be_number(self):
        with self.assertRaises(ExcelError):
            round('er', 1)

    @unittest.skip('This test fails.')
    def test_nb_digits_must_be_number(self):
        with self.assertRaises(ExcelError):
            round(2.323, 'ze')

    def test_positive_number_of_digits(self):
        self.assertEqual(xround(2.675, 2), 2.68)

    def test_negative_number_of_digits(self):
        self.assertEqual(xround(2352.67, -2), 2400)


class Test_Count(unittest.TestCase):
    def test_without_nested_booleans(self):
        range = Range('A1:A3', [1, 2, 'e'])

        self.assertEqual(count(range, True, 'r'), 3)

    def test_with_nested_booleans(self):
        range = Range('A1:A3', [1, True, 'e'])

        self.assertEqual(count(range, True, 'r'), 2)

    def test_with_text_representations(self):
        range = Range('A1:A3', [1, '2.2', 'e'])

        self.assertEqual(count(range, True, '20'), 4)


class Test_Countif(unittest.TestCase):
    def setUp(self):
        pass

    @unittest.skip('This test fails.')
    def test_argument_validity(self):
        range = Range('A1:A2', ['e', 1])

        with self.assertRaises(ExcelError):
            countif(range, '>=1')

    def test_countif_strictly_superior(self):
        range = Range('A1:A4', [7, 25, 13, 25])

        self.assertEqual(countif(range, '>10'), 3)

    def test_countif_strictly_inferior(self):
        range = Range('A1:A4', [7, 25, 13, 25])

        self.assertEqual(countif(range, '<10'), 1)

    def test_countif_superior(self):
        range = Range('A1:A4', [7, 25, 13, 25])

        self.assertEqual(countif(range, '>=10'), 3)

    def test_countif_inferior(self):
        range = Range('A1:A4', [7, 25, 10, 25])

        self.assertEqual(countif(range, '<=10'), 2)

    def test_countif_different(self):
        range = Range('A1:A4', [7, 25, 10, 25])

        self.assertEqual(countif(range, '<>10'), 3)

    def test_countif_with_string_equality(self):
        range = Range('A1:A4', [7, 'e', 13, 'e'])

        self.assertEqual(countif(range, 'e'), 2)

    def test_countif_regular(self):
        range = Range('A1:A4', [7, 25, 13, 25])

        self.assertEqual(countif(range, 25), 2)


class Test_Countifs(unittest.TestCase): # more tests might be welcomed
    def setUp(self):
        pass

    @unittest.skip('This test fails.')
    def test_countifs_not_associated(self):  # ASSOCIATION IS NOT TESTED IN COUNTIFS BUT IT SHOULD BE
        range1 = Range('A1:A4', [7, 25, 13, 25])
        range2 = Range('B2:B5', [100, 102, 201, 20])
        range3 = Range('C3:C7', [100, 102, 201, 20])

        with self.assertRaises(ExcelError):
            countifs(range1, 25, range2, ">100", range3, "<=100")

    def test_countifs_regular(self):
        range1 = Range('A1:A4', [7, 25, 13, 25])
        range2 = Range('B1:B4', [100, 102, 201, 20])

        self.assertEqual(countifs(range1, 25, range2, ">100"), 1)


class Test_Mod(unittest.TestCase):
    @unittest.skip('This test fails.')
    def test_first_argument_validity(self):
        with self.assertRaises(ExcelError):
            mod(2.2, 1)

    @unittest.skip('This test fails.')
    def test_second_argument_validity(self):
        with self.assertRaises(ExcelError):
            mod(2, 1.1)

    def test_output_value(self):
        self.assertEqual(mod(10, 4), 2)


class Test_Match(unittest.TestCase):
    def test_numeric_in_ascending_mode(self):
        range = Range('A1:A3', [1, 3.3, 5])
        # Closest inferior value is found
        self.assertEqual(match(5, range), 3)

    @unittest.skip('This test fails.')
    def test_numeric_in_ascending_mode_with_descending_array(self):
        range = Range('A1:A4', [10, 9.1, 6.23, 1])
        # Not ascending arrays raise exception
        with self.assertRaises(ExcelError):
            match(3, range)

    @unittest.skip('This test fails.')
    def test_numeric_in_ascending_mode_with_any_array(self):
        range = Range('A1:A4', [10, 3.3, 5, 2])
        # Not ascending arrays raise exception
        with self.assertRaises(ExcelError):
            match(3, range)

    def test_numeric_in_exact_mode(self):
        range = Range('A1:A3', [10, 3.3, 5.0])
        # Value is found
        self.assertEqual(match(5, range, 0), 3)

    @unittest.skip('This test fails.')
    def test_numeric_in_exact_mode_not_found(self):
        range = Range('A1:A4', [10, 3.3, 5, 2])
        # Value not found raises ExcelError
        with self.assertRaises(ExcelError):
            match(3, range, 0)

    def test_numeric_in_descending_mode(self):
        range = Range('A1:A3', [10, 9.1, 6.2])
        # Closest superior value is found
        self.assertEqual(match(8, range, -1), 2)

    @unittest.skip('This test fails.')
    def test_numeric_in_descending_mode_with_ascending_array(self):
        range = Range('A1:A4', [1, 3.3, 5, 6])
        # Non descending arrays raise exception
        with self.assertRaises(ExcelError):
            match(3, range, -1)

    @unittest.skip('This test fails.')
    def test_numeric_in_descending_mode_with_any_array(self):
        range = Range('A1:A4', [10, 3.3, 5, 2])
        # Non descending arrays raise exception
        with self.assertRaises(ExcelError):
            match(3, range, -1)

    def test_string_in_ascending_mode(self):
        range = Range('A1:A3', ['a', 'AAB', 'rars'])
        # Closest inferior value is found
        self.assertEqual(match('rars', range), 3)

    @unittest.skip('This test fails.')
    def test_string_in_ascending_mode_with_descending_array(self):
        range = Range('A1:A3', ['rars', 'aab', 'a'])
        # Not ascending arrays raise exception
        with self.assertRaises(ExcelError):
            match(3, range)

    @unittest.skip('This test fails.')
    def test_string_in_ascending_mode_with_any_array(self):
        range = Range('A1:A3', ['aab', 'a', 'rars'])

        with self.assertRaises(ExcelError):
            match(3, range)

    def test_string_in_exact_mode(self):
        range = Range('A1:A3', ['aab', 'a', 'rars'])
        # Value is found
        self.assertEqual(match('a', range, 0), 2)

    @unittest.skip('This test fails.')
    def test_string_in_exact_mode_not_found(self):
        range = Range('A1:A3', ['aab', 'a', 'rars'])
        # Value not found raises ExcelError
        with self.assertRaises(ExcelError):
            match('b', range, 0)

    def test_string_in_descending_mode(self):
        range = Range('A1:A3', ['c', 'b', 'a'])
        # Closest superior value is found
        self.assertEqual(match('a', range, -1), 3)

    @unittest.skip('This test fails.')
    def test_string_in_descending_mode_with_ascending_array(self):
        range = Range('A1:A3', ['a', 'aab', 'rars'])
        # Non descending arrays raise exception
        with self.assertRaises(ExcelError):
            match('a', range, -1)

    @unittest.skip('This test fails.')
    def test_string_in_descending_mode_with_any_array(self):
        ange = Range('A1:A3', ['aab', 'a', 'rars'])
        # Non descending arrays raise exception
        with self.assertRaises(ExcelError):
            match('a', ['aab', 'a', 'rars'], -1)

    def test_boolean_in_ascending_mode(self):
        range = Range('A1:A3', [False, False, True])
        # Closest inferior value is found
        self.assertEqual(match(True, range), 3)

    @unittest.skip('This test fails.')
    def test_boolean_in_ascending_mode_with_descending_array(self):
        range = Range('A1:A3', [True, False, False])
        # Not ascending arrays raise exception
        with self.assertRaises(ExcelError):
            match(False, range)

    @unittest.skip('This test fails.')
    def test_boolean_in_ascending_mode_with_any_array(self):
        range = Range('A1:A3', [False, True, False])
        # Not ascending arrays raise exception
        with self.assertRaises(ExcelError):
            match(True, range)

    def test_boolean_in_exact_mode(self):
        range = Range('A1:A3', [True, False, False])
        # Value is found
        self.assertEqual(match(False, range, 0), 2)

    @unittest.skip('This test fails.')
    def test_boolean_in_exact_mode_not_found(self):
        range = Range('A1:A3', [True, True, True])
        # Value not found raises ExcelError
        with self.assertRaises(ExcelError):
            match(False, range, 0)

    def test_boolean_in_descending_mode(self):
        range = Range('A1:A3', [True, False, False])
        # Closest superior value is found
        self.assertEqual(match(False, range, -1), 3)

    @unittest.skip('This test fails.')
    def test_boolean_in_descending_mode_with_ascending_array(self):
        range = Range('A1:A3', [False, False, True])
        # Non descending arrays raise exception
        with self.assertRaises(ExcelError):
            match(False, range, -1)

    @unittest.skip('This test fails.')
    def test_boolean_in_descending_mode_with_any_array(self):
        range = Range('A1:A3', [False, True, False])
        with self.assertRaises(ExcelError):
            match(True, range, -1)
